#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct 19 15:36:42 2020

@author: rakiki
"""
import numpy as np
import rasterio
# dataset construction
def pointCube(dim1_left, dim1_right, dim2_left, dim2_right
              , alt_min, alt_max
              , grid_len, num_layers):
    
    '''
    Construct a 3D meshgrid 
    
    Args:
    dim1, dim2 can either be lon, lat (or lat, lon /order is irrelevant)
                            or row, col (or col, row /order is irrelevant )
    dim1/2_left/right: the limits of the dimension range
    alt_min/max: the limits of the altitude range
    grid_len: the number of points to be taken in dim1/2
    num_layers: the number of points to be taken in the altitude range
    Returns: 
    dim1_list, dim2_list, alt_list: a 1D numpy array of values in each dimension
    '''
    # meshgrid 
    dim1_rg = np.linspace(dim1_left,dim1_right, grid_len)
    dim2_rg = np.linspace(dim2_left, dim2_right, grid_len)
    alt_rg = np.linspace(alt_min,alt_max, num_layers)
    dim1_grd, dim2_grd, alt_grd = np.meshgrid(dim1_rg, dim2_rg, alt_rg)
    dim2_list = dim2_grd.ravel()
    dim1_list = dim1_grd.ravel()
    alt_list = alt_grd.ravel()
    return dim1_list, dim2_list, alt_list

def read_dem(dem_path):
    '''
    Reads the dem that covers the area of intererst, 
    geotiff, from the disk
    dem_path: the path to the dem file
    Returns:
        demdb: rasterio closed dataset
        demdata: numpy.2D array containing the dem 
    '''
    with rasterio.open(dem_path) as demdb:
        demdata = demdb.read(1)
    return demdb, demdata

def getDataset_projection(demdb, demdata , grid_len, num_layers, projectionFunc
                          , train = True, margin = 0.2, **kwargs):
    '''
    Computes 3D pts + 2D correspondence for train or test set
    Args: 
    demdb: rasterio db of the geotiff dem on the area of interest
    demdata: the data of the dem 
    grid_len: the len of the grid in the two lon, lat dimensions 
    num_layers: the number of alt layers
    projectionFunc: projection function (lon, lat, alt) -> (col, row)
    train: if True, returns a training grid else returns a test grid (shifted by half a step from the train grid)
    margin: the safety margin to apply to the altitude bounds when constructing the grid
    kwargs: dict of params to pass to projection function 
    Returns: 
    input_locs: [lon, lat, alt] array
    target: [col, row] array
    '''
    lon_left, lat_top = demdb.transform * (0,0)
    lon_right,lat_bottom =  demdb.transform * (demdb.width -1 ,demdb.height -1 )
    mask = (demdata != demdb.nodata)
    alt_min = np.nanmin(demdata[mask])
    alt_max = np.nanmax(demdata[mask])
    alt_margin = np.round((alt_max - alt_min) * margin)
    alt_min -= alt_margin
    alt_max += alt_margin
    if train:
        lon,lat, alt = pointCube(lon_left, lon_right, lat_top, 
                                           lat_bottom, alt_min , 
                                           alt_max, grid_len, num_layers, 
                                           )
    else:
        lon_stp  = (lon_right - lon_left)/(2 * (grid_len - 1 ) ) 
        lat_stp = (lat_bottom - lat_top)/(2 * (grid_len - 1 ) ) 
        alt_stp = np.round((alt_max - alt_min)/(2 * (num_layers - 1 ) ) )  
        lon, lat, alt = pointCube(lon_left + lon_stp , lon_right + lon_stp, lat_top + lat_stp, 
                                                lat_bottom + lat_stp, alt_min + alt_stp, alt_max + alt_stp
                                                , grid_len, num_layers, 
                                               )
    col, row = projectionFunc( lon = lon , lat = lat
                              , alt = alt , **kwargs)
    input_locs = np.vstack((lon, lat, alt)).T
    target =  np.vstack((col,row)).T
    return input_locs, target

def getDataset_localization(demdb, demdata, grid_len, num_layers
                            , im_size , localizationFunc, train = True, margin = 0.2 
                            , **kwargs ):
    '''
    Computes 3D pts + 2D correspondence for train or test set
    
    Args: 
    demdb: rasterio db of the geotiff dem on the area of interest
    demdata: the data of the dem
    grid_len: the len of the grid in the two lon, lat dimensions 
    num_layers: the number of alt layers
    im_size: tuple(height, width) of the image
    localizationFunc: localization function (col, line, alt) -> (lon, lat)
    train: if True, returns a training grid else returns a test grid (shifted by half a step from the train grid)
    margin: the safety margin to apply to the bounds of the image dimension and the altitude bounds when constructing the grid
    kwargs: localization function additional arguments
    Returns: 
    input_locs: [lon, lat, alt] array
    target: [col, row] array
    '''
    # line, col limits
    lines = im_size[0]
    l_margin = np.round(margin * lines)
    columns = im_size[1]
    c_margin = np.round(margin * columns)

    # alt limits, use preexisting demdb, demdata
    mask = (demdata != demdb.nodata)
    alt_min = np.nanmin(demdata[mask]) 
    alt_max = np.nanmax(demdata[mask])
    alt_margin = np.round((alt_max - alt_min) * margin)

    if train:
        c,l, alt = pointCube(-c_margin, columns + c_margin, -l_margin, 
                            lines + l_margin, alt_min - alt_margin , alt_max + alt_margin,
                             grid_len, num_layers)
    else:
        c_stp  = (columns + 2 * c_margin)/(2 * (grid_len - 1 ) ) 
        l_stp = (lines + 2 * l_margin)/(2 * (grid_len - 1 ) ) 
        alt_stp = np.round((alt_max - alt_min + 2 * alt_margin)/(2 * (num_layers - 1 ) ) )  
        c, l, alt = pointCube(-c_margin + c_stp , columns + c_margin + c_stp, -l_margin + l_stp, 
                                                lines + l_margin + l_stp, alt_min - alt_margin + alt_stp
                                                  , alt_max + alt_margin + alt_stp, 
                                                grid_len, num_layers)
    lon , lat = localizationFunc(col = c, line = l, alt = alt, **kwargs)
    input_locs = np.vstack((lon, lat, alt)).T
    target =  np.vstack((c,l)).T
    return input_locs, target


from typing import Optional
import numpy as np
from rpcfit import Lcurve
from rpcm import rpc_model 

def poly_vect(x, y, z):
    """
    Returns evaluated polynomial vector without the first constant term equal to 1,
    using the order convention defined in rpc_model.apply_poly
    """
    return np.array([y, x, z, 
                     y*x, y*z, x*z, y*y, x*x, z*z, 
                     x*y*z, y*y*y, y*x*x, y*z*z, y*y*x, x*x*x, x*z*z, y*y*z, x*x*z, z*z*z])

def normalize_target(rpc, target):
    """
    Normalize in image space
    """
    target_norm = np.vstack(((target[:, 0] - rpc.col_offset) / rpc.col_scale,
                             (target[:, 1] - rpc.row_offset) / rpc.row_scale)).T
    return target_norm

def normalize_input_locs(rpc, input_locs):
    """
    Normalize in world space
    """
    input_locs_norm = np.vstack(((input_locs[:, 0] - rpc.lon_offset) / rpc.lon_scale,
                                 (input_locs[:, 1] - rpc.lat_offset) / rpc.lat_scale,
                                 (input_locs[:, 2] - rpc.alt_offset) / rpc.alt_scale)).T
    return input_locs_norm    

def update_rpc(rpc, x, orientation = "projection"):
    """
    Update rpc coefficients
    orientation: if "projection" assume the coefficients are for row and col
    otherwise for lon and lat
    """
    if orientation == "projection": 
        rpc.row_num, rpc.row_den = x[:20], x[20:40]
        rpc.col_num, rpc.col_den = x[40:60], x[60:]
    else: 
        rpc.lat_num, rpc.lat_den = x[:20], x[20:40]
        rpc.lon_num, rpc.lon_den = x[40:60], x[60:]
    return rpc

def calculate_RMSE_row_col(rpc, input_locs, target):
    """
    Calculate MSE & RMSE in image domain
    """
    col_pred, row_pred = rpc.projection(lon=input_locs[:,0], lat=input_locs[:,1], alt=input_locs[:,2])
    MSE_col, MSE_row = np.mean((np.hstack([col_pred.reshape(-1, 1), row_pred.reshape(-1, 1)]) - target) ** 2, axis=0)
    MSE_row_col = np.mean([MSE_col, MSE_row]) # the number of data is equal in MSE_col and MSE_row
    RMSE_row_col = np.sqrt(MSE_row_col)
    return RMSE_row_col

def evaluate(rpc, input_locs, target):
    '''
    Args: 
    rpc: RPCModel obj
    input_locs: Nx3 (lon, lat, alt)
    target: Nx2 (col, row)
    Returns: 
    RMSE (col, row) 
    max_err tuple (col, row) ,
    planimetry Nx1 (error in px))
    '''
    col_pred, row_pred = rpc.projection(lon=input_locs[:,0], lat=input_locs[:,1], alt=input_locs[:,2])
    error = np.hstack([col_pred.reshape(-1, 1), row_pred.reshape(-1, 1)]) - target
    max_err = np.amax(np.abs(error), axis = 0)
    RMSE = np.sqrt(np.mean(error**2, axis = 0))
    planimetry = np.linalg.norm(error, axis = 1)
    return RMSE , max_err, planimetry

def evaluate_localization(rpc, input_locs, target):
    '''
    Args: 
    rpc: RPCModel obj
    input_locs: Nx3 (lon, lat, alt)
    target: Nx2 (col, row)
    Returns:
    RMSE (lon, lat) 
    max_err tuple (lon, lat) ,
    planimetry Nx1 (error))
    '''
    lon_pred, lat_pred = rpc.localization(col=target[:,0], row=target[:,1], alt=input_locs[:,2])
    error = np.hstack([lon_pred.reshape(-1, 1), lat_pred.reshape(-1, 1)]) - input_locs[:,:2]
    max_err = np.amax(np.abs(error), axis = 0)
    RMSE = np.sqrt(np.mean(error**2, axis = 0))
    planimetry = np.linalg.norm(error, axis = 1)
    return RMSE , max_err, planimetry

def weighted_lsq(rpc_to_calibrate, target, input_locs, h=1e-3, tol=1e-2, max_iter=20):
    """
    Regularized iterative weighted least squares for calibrating rpc.
    
    Args: 
        max_iter : maximum number of iterations
        h : regularization parameter
        tol : tolerance criterion on improvment of RMSE over iterations
        
    Warning: this code is to be employed with the rpc_model defined in s2p
    """
    reg_matrix = (h ** 2) * np.eye(39)  # regularization matrix
    target_norm = normalize_target(rpc_to_calibrate, target)  # col, row
    input_locs_norm = normalize_input_locs(rpc_to_calibrate, input_locs)  # lon, lat, alt

    # define C, R and M
    C, R = target_norm[:, 0][:, np.newaxis], target_norm[:, 1][:, np.newaxis]
    lon, lat, alt = input_locs_norm[:,0], input_locs_norm[:,1], input_locs_norm[:,2]
#    col, row = target_norm[:,0][:, np.newaxis], target_norm[:,1][:, np.newaxis]
    vect = poly_vect(x=lat, y=lon, z=alt).T
    MC = np.hstack([np.ones((lon.shape[0], 1)), vect, -C * vect])
    MR = np.hstack([np.ones((lon.shape[0], 1)), vect, -R * vect])
       
    # calculate direct solution
    JR = np.linalg.inv(MR.T @ MR) @ (MR.T @ R)
    JC = np.linalg.inv(MC.T @ MC) @ (MC.T @ C)

    # update rpc and get error
    coefs = np.vstack([JR[:20], 1, JR[20:], JC[:20], 1, JC[20:]]).reshape(-1)
    rpc_to_calibrate = update_rpc(rpc_to_calibrate, coefs)
    RMSE_row_col = calculate_RMSE_row_col(rpc_to_calibrate, input_locs, target)
    for n_iter in range(1, max_iter+1):
        WR2 = np.diagflat(1 / ((MR[:, :20] @ coefs[20:40]) ** 2))  # diagonal matrix with 1 / denom ** 2
        JR_iter = np.linalg.inv((MR.T @ WR2 @ MR) + reg_matrix) @ (MR.T @ WR2 @ R)
        WC2 = np.diagflat(1 / ((MC[:, :20] @ coefs[60:80]) ** 2))  # diagonal matrix with 1 / denom ** 2
        JC_iter = np.linalg.inv((MC.T @ WC2 @ MC) + reg_matrix) @ (MC.T @ WC2 @ C)

        # update rpc and get error
        coefs = np.vstack([JR_iter[:20], 1, JR_iter[20:], JC_iter[:20], 1, JC_iter[20:]]).reshape(-1)
        rpc_to_calibrate = update_rpc(rpc_to_calibrate, coefs)
        RMSE_row_col_prev = RMSE_row_col
        RMSE_row_col = calculate_RMSE_row_col(rpc_to_calibrate, input_locs, target)
        # check convergence
        if np.abs(RMSE_row_col_prev - RMSE_row_col) < tol:
            break
    
    return rpc_to_calibrate
    
def get_mats(rpc_to_calibrate, target, input_locs, orientation = "projection"):
    '''
    Computes the design matrices used in the least squares fit
    Args: 
        rpc_to_calibrate: RPCModel instance, should contain normalization coefficients
        input_locs: Nx3 (lon, lat, alt)
        target: Nx2 (col, row)
        orientation: "projection" returns matrices for projection rpc fitting
                     "localization" returns matrices for localization rpc fitting
    Returns: 
        MR, MC, A: design matrices for row, col, joint problem estimation respectively
        R, C, l: the right hand side term for row, col, joint problem estimation respectively

    '''
    target_norm = normalize_target(rpc_to_calibrate, target)  # col, row
    input_locs_norm = normalize_input_locs(rpc_to_calibrate, input_locs)  # lon, lat, alt
    col , row = target_norm[:, 0], target_norm[:, 1]
    lon, lat, alt = input_locs_norm[:, 0], input_locs_norm[:, 1], input_locs_norm[:, 2]
    if orientation == "projection": 
        # define C, R  and x, y, z
        C, R = col[:, np.newaxis], row[:, np.newaxis]
        x, y, z = lat, lon, alt # same order as RPCModel 
    else: 
        # define C, R and x, y, z 
        C, R = lon[:, np.newaxis], lat[:, np.newaxis]
        x, y, z = row, col, alt # same order as RPCModel 
    vect = poly_vect(x=x, y=y, z=z).T
    num_samp = C.shape[0]
    MC = np.hstack([np.ones((num_samp, 1)), vect, -C * vect])
    MR = np.hstack([np.ones((num_samp, 1)), vect, -R * vect])
    A = np.zeros((2 * num_samp, 78))
    A[:num_samp, :39] = MR
    A[num_samp:, 39:] = MC
    l = np.vstack([R, C])
    return MR, R, MC, C, A, l
    
def compute_rmse(X, y, theta, scale = (1,)):
    '''
    computes rmse without rpc model
    Args: 
        X: design mat
        y: right side term
        theta: current solution
        scale: y normalization scale
    Returns: 
        RMSE
    '''
    weights = get_weight(X, theta)
    # check scale
    if theta.size == 39: 
        scaling = scale[0] * np.ones((X.shape[0], 1))
    elif theta.size == 78:
        num_samp = X.shape[0]//2
        scaling = np.vstack([scale[0] * np.ones((num_samp, 1)), 
                             scale[1] * np.ones((num_samp, 1))])
    RMSE = np.sqrt(np.mean((scaling*weights*(X @ theta - y))**2))
    return RMSE, weights

def get_weight(X, theta):
    '''
    Computes the weights from the desing matrix and current solution
    Args: 
        X: design matrix
        theta: current solution
    Returns: 
        weights
    '''
    if theta.size == 39:
        # Seperately solving row and col
        weights = 1/ ( X[:,:20] @ np.vstack([ 1, theta[20:39]]))
    elif theta.size == 78:
        # jointly solving row and col
        num_samp = int(X.shape[0]//2)
        R = 1 / ( X[:num_samp, :20] @ np.vstack([ 1, theta[20:39]]))
        C = 1 / ( X[num_samp:, 39:59] @ np.vstack([ 1, theta[59:78]]) )
        weights = np.vstack([R.reshape(-1,1), C.reshape(-1, 1)])
    return weights.reshape(-1,1)
  
def wlsq_svd(X , y, tol=1e-2, max_iter=20, h = None, plot = False, scale = (1,)):
    '''
    iterative weighted least squares, initial solution and regularization parameter 
    computed by means of Lcurve heuristic, if regularization parameter not given
    Then, joint row and col iterative weighted regularized least squares optim
    based on svd decomposition
    each iteration performs and svd decomposition to find the solution
    Args: 
        X: design matrix
        y: right side term
        max_iter : maximum number of iterations
        tol : tolerance criterion on improvment of RMSE over iterations
        h: regularization parameter, if None, compute by Lcurve
        plot: whether to plot Lcurve, ignore if Lcurve is not used
        scale: scale used to normalize the right side coordinate
                tuple (scale_row, scale_col) when solving jointly, otherwise 
                tuple (scale_row, ) or tuple (scale_col, )
    Returns: 
        solution vector, norm_res, norm_sol
    '''    
    
    if h is None: 
        # first iteration, svd solver
        h, theta = Lcurve.l_curve(X, y, plotit=plot)
        norm_sol_best = np.linalg.norm(theta)
        norm_res_best = np.linalg.norm(X @ theta - y)
    else: 
        norm_res_best, norm_sol_best, theta = Lcurve.LSQ_iter(X, y , h)
    theta = theta.reshape((-1,1))
    # get error
    RMSE, weights = compute_rmse(X, y, theta, scale)
    theta_best = theta
    RMSE_best = RMSE
    
    # weighted least squares
    for n_iter in range(1, max_iter+1):
            # iterate, svd solver
            norm_res , norm_sol , theta_iter = Lcurve.LSQ_iter(weights*X, weights*y , h)
            theta_iter = theta_iter.reshape((-1,1))
            # store previous rmse 
            RMSE_prev = RMSE
            # get rmse and weights
            RMSE, weights = compute_rmse(X, y, theta_iter, scale)
            # store best solution
            if RMSE < RMSE_best: 
                theta_best = theta_iter
                RMSE_best = RMSE
                norm_res_best = norm_res
                norm_sol_best = norm_sol
            # check convergence
            if np.abs(RMSE_prev - RMSE) < tol:
                break
    return theta_best, norm_res_best, norm_sol_best, h

def solve_lc(X, y, tol=1e-2, max_iter=20, method = "initLcurve", plot = False, scale = (1,)):
    '''
    Solves argmin(theta) || X.theta - y ||^2 + h^2 ||theta||^2
    h is the regularization parameter found automatically by Lcurve
    Uses weighted least squares with svd solver
    Args:
        X: design matrix
        y: right side term
        max_iter : maximum number of iterations
        tol : tolerance criterion on improvment of RMSE over iterations
        method: the method of finding the regularization h automatically
                'initLcurve' will do a standard tikhonov Lcurve without weights
                to find the h parameter, uses closed form formulas
                'discreteLcurve' solves the problem N times with N regularization
                parameters, Lcurve is then fitted on the discrete data points
        scale: scale used to normalize the right side coordinate
                tuple (scale_row, scale_col) when solving jointly, otherwise 
                tuple (scale_row, ) or tuple (scale_col, )
    Returns: 
        theta: the solution vector for the optimal regularization parameter
    '''
    methods = ["initLcurve", "discreteLcurve"]
    if not method in methods: 
        method = "initLcurve"
    if method == "initLcurve": 
        theta, _, _, regu = wlsq_svd(X , y, tol=tol, max_iter=max_iter, plot = plot, scale = scale)
    else: 
        npoints = 100
        eta = np.zeros((npoints,1))
        rho = np.zeros((npoints,1)) 
        thetas = np.zeros((npoints,X.shape[1]))
        reg_param = Lcurve.get_reg_param(X, npoints, truncate = False)
        # solve problem for each regularization param
        for i, h in enumerate(reg_param):
            # the solver is a weighted iterative solver
            sol, norm_res, norm_sol, _ = wlsq_svd(X, y, tol = tol, max_iter = max_iter, h = h, scale = scale)
            thetas[i] = sol.ravel()
            eta[i] = norm_sol
            rho[i] = norm_res
        # find the corner of the Lcurve
        cid = Lcurve.find_corner(eta, rho, plot = plot)
        theta = thetas[cid].reshape(-1,1)
        regu = reg_param[cid]
    return theta, regu

def solve(X, y, tol=1e-2, max_iter=20, method = "initLcurve", plot = False, scale = (1,) ):
    '''
    Solves argmin(theta) || X.theta - y ||^2 + ||theta||^2 
    with ICCV(iteration by correcting characteristic value)
    Initial value determined by Lcurve
    Args:
        X: design matrix
        y: right side term
        max_iter : maximum number of iterations
        tol : tolerance criterion on improvment of RMSE over iterations
        method: the method of finding the regularization h automatically
                'initLcurve' will do a standard tikhonov Lcurve without weights
                to find the h parameter, uses closed form formulas
                'discreteLcurve' solves the problem N times with N regularization
                parameters, Lcurve is then fitted on the discrete data points
        scale: scale used to normalize the right side coordinate
                tuple (scale_row, scale_col) when solving jointly, otherwise 
                tuple (scale_row, ) or tuple (scale_col, )
    Returns: 
        theta: the solution vector for the optimal regularization parameter
    '''
    theta, regu = solve_lc(X, y, tol, max_iter, method, plot, scale )
    theta = theta.reshape((-1,1))
    # get error
    RMSE, weights = compute_rmse(X, y, theta, scale)
    theta_best = theta
    RMSE_best = RMSE
    # ICCV (iteration by correcting characteristic value)
    Id = np.eye(X.shape[1])
    for n_iter in range(1, max_iter+1):
            # iterate
            diag = weights*weights
            theta = np.linalg.inv(X.T @ (diag * X) + Id) @ (X.T @ (diag * y) + theta)
            theta = theta.reshape((-1,1))
            # store previous rmse 
            RMSE_prev = RMSE
            # get rmse and weights
            RMSE, weights = compute_rmse(X, y, theta, scale)
            # store best solution
            if RMSE < RMSE_best: 
                theta_best = theta
                RMSE_best = RMSE
            # check convergence
            if np.abs(RMSE_prev - RMSE) < tol:
                break
    return theta_best, regu

def empty_rpc():
    '''
    Creates an empty rpc instance
    '''
    d = {}
    listkeys = ['LINE_OFF','SAMP_OFF','LAT_OFF','LONG_OFF','HEIGHT_OFF'
           ,'LINE_SCALE','SAMP_SCALE','LAT_SCALE','LONG_SCALE','HEIGHT_SCALE',
           'LINE_NUM_COEFF','LINE_DEN_COEFF','SAMP_NUM_COEFF','SAMP_DEN_COEFF']
    for key in listkeys:
        d[key]= '0'
    return rpc_model.RPCModel(d)

def scaling_params(vect):
    '''
    returns scale, offset based 
    on vect min and max values
    '''
    min_vect = min(vect)
    max_vect = max(vect)
    mid = (max_vect - min_vect)/2 
    return mid, min_vect + mid

def init_rpc(target, input_locs):
    '''
    Initialize an rpc instance on (row, col , lon, lat, h) grids
    by initializing the scale and offset parameters
    '''
    rpc_init = empty_rpc()
    rpc_init.row_scale,rpc_init.row_offset  = scaling_params(target[:,1])
    rpc_init.col_scale,rpc_init.col_offset  = scaling_params(target[:,0])
    rpc_init.lat_scale,rpc_init.lat_offset  = scaling_params(input_locs[:,1])
    rpc_init.lon_scale,rpc_init.lon_offset  = scaling_params(input_locs[:,0])
    rpc_init.alt_scale,rpc_init.alt_offset  = scaling_params(input_locs[:,2])
    return rpc_init
      
def calibrate_rpc( target, input_locs, separate = True, tol=1e-2
                  , max_iter=20, method = "initLcurve", plot = False, 
                  orientation = "projection", get_log = False,
                  init: Optional[rpc_model.RPCModel] = None):
    '''
    fits the coefficients of an RPCModel instance with
    regularized weighted least squares on 3D -> 2D grid correspondence
    Args:
        target: Nx2 (col, row)
        input_locs: Nx3 (lon, lat, alt)
        separate: if True, fit the row and col projection separately
        max_iter : maximum number of iterations
        tol : tolerance criterion on improvment of RMSE over iterations
        method: the method of finding the regularization h automatically
                'initLcurve' will do a standard tikhonov Lcurve without weights
                to find the h parameter, uses closed form formulas
                'discreteLcurve' solves the problem N times with N regularization
                parameters, Lcurve is then fitted on the discrete data points
                using 'discreteLcurve' is discouraged as it is costly
        plot: if True, plot Lcurves
        orientation: "projection" to fit projection rpcs
                     "localization" to fit localization rpcs
                     "projloc" to fit projection and localization rpcs
    Returns: 
        rpc_calib: calibrated RPCModel instance
    '''
    # if not given, initialize an empty rpc instance
    rpc_calib = init or init_rpc(target, input_locs)
    if orientation == "projloc":
        orientlist = ["projection", "localization"]
    else: 
        orientlist = [orientation]
    logger = {}
    for orient in orientlist:
        # get the matrices of the problem
        MR, R , MC, C, A, l = get_mats(rpc_calib, target, input_locs, orientation = orient)     
        # solve jointly or separately
        if separate: 
            thetaR, hr = solve(MR,R,tol,max_iter,method,plot,(rpc_calib.row_scale,))
            thetaC, hc = solve(MC,C,tol,max_iter,method,plot,(rpc_calib.col_scale,))
            logger[orient] = {'hr':hr, 'hc': hc}
        else: 
            theta, h = solve(A,l,tol,max_iter,method,plot, (rpc_calib.row_scale
                                                         ,rpc_calib.col_scale))
            logger[orient] = {'h': h}
            thetaR = theta[:39]
            thetaC = theta[39:]
        coeffs = np.vstack([thetaR[:20], 1, thetaR[20:39],
                            thetaC[:20], 1, thetaC[20:39]]).reshape(-1)
        rpc_calib = update_rpc(rpc_calib, coeffs, orientation = orient)
    if get_log:    
        return rpc_calib, logger
    else:
        return rpc_calib



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Sep 23 13:08:19 2020

@author: rakiki

Inspired by lcurve implementation from https://github.com/eric-brandao/lcurve
"""
from scipy import interpolate,optimize,linalg
import numpy as np
import matplotlib.pyplot as plt

### Standard tikhonov Lcurve functions
def curvature(lambd, sig, beta, xi, rhoLS2):
    '''
    Computes the curvature.(-1) of the standard tikhonov Lcurve
    Args: 
        lambd: 1D np array of regularization parameter values
        sig: 1D np array of singular values for design matrix
        beta: projection of right equation term onto left singular vectors
        xi: beta / sig with safedivision
        rhoLS2: the orthogonal residual 
    Returns : 
        The negative of the curvature ( to minimize it instead of maximize)
    '''
    # Initialization.
    phi = np.zeros(lambd.shape)
    dphi = np.zeros(lambd.shape)
    psi = np.zeros(lambd.shape)
    dpsi = np.zeros(lambd.shape)
    eta = np.zeros(lambd.shape)
    rho = np.zeros(lambd.shape)

    # Compute some intermediate quantities.
    for jl, lam in enumerate(lambd):
        f  = np.divide((sig ** 2), (sig ** 2 + lam ** 2)) 
        cf = 1 - f 
        eta[jl] = np.linalg.norm(f * xi) 
        rho[jl] = np.linalg.norm(cf * beta)
        f1 = -2 * f * cf / lam 
        f2 = -f1 * (3 - 4*f)/lam
        phi[jl]  = np.sum(f*f1*np.abs(xi)**2) 
        psi[jl] = np.sum(cf*f1*np.abs(beta)**2)
        dphi[jl] = np.sum((f1**2 + f*f2)*np.abs(xi)**2)
        dpsi[jl] = np.sum((-f1**2 + cf*f2)*np.abs(beta)**2) 
    rho = np.sqrt(rho ** 2 + rhoLS2)

    # Now compute the first and second derivatives of eta and rho
    # with respect to lambda;
    deta  =  np.divide(phi, eta) 
    drho  = -np.divide(psi, rho)
    ddeta =  np.divide(dphi, eta) - deta * np.divide(deta, eta)
    ddrho = -np.divide(dpsi, rho) - drho * np.divide(drho, rho)

    # Convert to derivatives of log(eta) and log(rho).
    dlogeta  = np.divide(deta, eta)
    dlogrho  = np.divide(drho, rho)
    ddlogeta = np.divide(ddeta, eta) - (dlogeta)**2
    ddlogrho = np.divide(ddrho, rho) - (dlogrho)**2
    # curvature.
    curv = - np.divide((dlogrho * ddlogeta - ddlogrho * dlogeta),
        (dlogrho**2 + dlogeta**2)**(1.5))
    return curv

def safedivision(num, denum):
    '''
    Computes the elementwise division of vec1 by vec2
    if vec2[i] = 0, result[i] = 0
    Args: 
        num, denum: np.array
    Returns: 
        num / denum, with 0 where denum = 0
    '''
    denum = denum.reshape(num.shape)
    result = np.zeros_like(num)
    mask = np.where(denum!= 0)
    result[mask] = num[mask]/denum[mask]
    return result

def l_corner(reg_param,u,sig,bm, plotit = False):
    '''
    computes the corner of the L-curve for standard tikhonov
    Inputs:
        reg_param - computed in l_curve function
        u left side matrix computed from svd (size: Nm x Nu) 
        sig is the singular value vector of A
        bm is the measured results
        plotit : whether to plot the curvature
   Returns: 
        reg_c: best regularization parameter, not necessarily in reg_param
        curv_id: id of regularization parameter closest to reg_c
    '''
    # Set default parameters for treatment of discrete L-curve.
    Nm, Nu = u.shape
    npoints = len(reg_param)
    beta = (u.T @ bm).reshape(-1,1)
    # ortho resid
    b0 = bm - u.dot(beta)
    rhoLS2 = np.linalg.norm(b0)**2
    beta = beta.ravel()
    xi = safedivision(beta, sig)
    # Call curvature calculator
    curv = curvature(reg_param, sig, beta, xi, rhoLS2) 
    # Minimize on interval
    curv_id = np.argmin(curv)
    x1 = reg_param[int(np.amin([curv_id + 1, npoints - 1 ]))]
    x2 = reg_param[int(np.amax([curv_id - 1, 0]))]
    if plotit: 
        # TODO fix plotting
        plt.figure()
        plt.title('curvature')
        plt.xlabel('regularization')
        plt.ylabel('curvature')
        plt.plot(reg_param, -curv, marker = '.')
        plt.scatter(reg_param[curv_id], -curv[curv_id], marker = 'x', color = 'r')
        plt.show()
    reg_c = optimize.fminbound(curvature, x1, x2, args = (sig, beta, xi, rhoLS2), full_output=False, disp=False)
    kappa_max = - curvature(reg_c, sig, beta, xi, rhoLS2)# Maximum curvature.
    if kappa_max[0] < 0:
        # in case the curvature is always negative ( concave )
        # take values for the smallest regularization param
        reg_c = reg_param[-1]
    return reg_c[0], curv_id
    

def l_curve(A, bm, plotit = False):
    '''
    Plot the L-curve and find its "corner" for standard tikhonov 
    Inputs:
        A: design matrix, is of size Nm x Nu,
        where Nm are the number of measurements and Nu the number of unknowns
        bm: your measurement vector (size: Nm x 1)
        plotit: plot the l curve and the curvature
    Returns: 
        lam_opt: the optimal regularization parameter
    '''
    u, sig , v = linalg.svd(A, full_matrices=False)
    # Set defaults.
    npoints = 200  # Number of points on the L-curve
    smin_ratio = 16*np.finfo(float).eps  # Smallest regularization parameter.
    # Initialization.
    Nm, Nu = u.shape
    beta = (u.T @ bm ).reshape(-1,1)
    b0 = bm - u.dot(beta)
    rhoLS2 = np.linalg.norm(b0)**2
    beta = beta.ravel()
    xi = safedivision(beta, sig)

    eta = np.zeros((npoints,1))
    rho = np.zeros((npoints,1)) 
    reg_param = np.zeros((npoints,1))
    s2 = sig ** 2
    # reg_param list from smallest singular val to biggest
    reg_param[-1] = np.amax([sig[-1], sig[0]*smin_ratio])
    ratio = (sig[0]/reg_param[-1]) ** (1/(npoints-1))
    for i in np.arange(start=npoints-2, step=-1, stop = -1):
        reg_param[i] = ratio*reg_param[i+1]
    for i in np.arange(start=0, step=1, stop = npoints):
        f = s2 / (s2 + reg_param[i] ** 2) # all filter coefs
        eta[i] = np.linalg.norm(f * xi)
        rho[i] = np.linalg.norm((1-f) * beta)
    if (Nm > Nu):
        rho = np.sqrt(rho ** 2 + rhoLS2)
    # Compute the corner of the L-curve (optimal regularization parameter)
    lam_opt, c_id = l_corner(reg_param,u,sig,bm, plotit)
    # Compute the optimal solution 
    f = s2 / (s2 + lam_opt** 2) # all filter coefs
    sol_opt = np.sum((f * xi).reshape(-1,1) * v, axis = 0).reshape(-1,1)
    # want to plot the L curve?
    if plotit:
        # TODO fix plotting
        plt.figure()
        plt.loglog(rho ,eta , marker = '.')
        plt.scatter(rho[c_id],eta[c_id], marker = 'X', color = 'r')
        plt.xlabel('Residual norm ||Ax - b||')
        plt.ylabel('Solution norm ||x||')
        plt.show()

    return lam_opt, sol_opt


def LSQ_iter(A, l, h):
    '''
    returns the lsq solution and norm, and resid norm of
    min || AJ - l ||**2 + h**2 ||J||**2
    with svd decomposition
    '''
    u, sig, v = linalg.svd(A, full_matrices=False)
    bm = l
    # Initialization.
    Nm, Nu = u.shape
    beta = (u.T @ bm ).reshape(-1,1)
    b0 = bm - u.dot(beta)
    rhoLS2 = np.linalg.norm(b0)**2
    beta = beta.ravel()
    xi = safedivision(beta, sig)
    reg_param = h
    s2 = sig ** 2
    f = s2 / (s2 + reg_param ** 2) # all filter coefs
    solution = np.sum((f * xi).reshape(-1,1) * v, axis = 0).reshape(-1,1)
    eta = np.linalg.norm(f * xi)
    rho = np.linalg.norm((1-f) * beta)
    if (Nm > Nu ):
        rho = np.sqrt(rho ** 2 + rhoLS2)
    return rho, eta, solution

def LSQ_iter_bis(X, y, h):
    '''
    returns the lsq solution and norm, and resid norm of
    min || AJ - l ||**2 + h**2 ||J||**2
    with svd decomposition
    Adapted from sklearn
    '''
    U, s, Vt = linalg.svd(X, full_matrices=False)
    idx = s > 1e-15  # same default value as scipy.linalg.pinv
    s_nnz = s[idx][:, np.newaxis]
    UTy = np.dot(U.T, y)
    d = np.zeros((s.size,1), dtype=X.dtype)
    d[idx] = s_nnz / (s_nnz ** 2 + h**2)
    d_UT_y = d * UTy
    sol =  np.dot(Vt.T, d_UT_y)
    residue = np.linalg.norm(X.dot(sol) - y)
    norm = np.linalg.norm(sol)
    return residue, norm, sol


def get_reg_param(X, npoints, truncate = False):
    '''
    Generates the list of regularization parameters on which to evaluate 
    the L curve
    Args: 
        X: the design matrix of the LSQ problem
        npoints: the number of points on the lcurve, ie the length of the 
                regularization parameters list
        truncate: if False, regularization goes from sig_min to sig_max in a geometric manner
                    if True, we just consider the principal singular value ( is not really stable)
    Returns: 
        The regularization parameters list
    '''
    reg_param = np.zeros((npoints,1))
    _, sig, _ = linalg.svd(X, full_matrices=False)
    smin_ratio = 100*np.finfo(float).eps 
    #TODO check if this strategy is always valid
    if truncate:
        smax = sig[np.argmax(sig[:-1]/sig[1:]) + 1]
    else: 
        smax = sig[0]
    # smallest regularization parameter
    reg_param[-1] = np.amax([sig[-1], smax*smin_ratio])
    ratio = (smax/reg_param[-1]) ** (1/(npoints-1))
    for i in np.arange(start=npoints-2, step=-1, stop = -1):
        reg_param[i] = ratio*reg_param[i+1]
    return reg_param

def find_corner(eta, rho, plot = False):
    '''
    Finds the corner of the Lcurve, based on spline interpolation
    Args: 
        eta: the list of solution norms
        rho: the list of residu norms
        plot: Whether to plot the Lcurve and the curvature
    Returns: 
        id of the corner of the Lcurve
    '''
    deg = 2 # degree of local smoothing polynomial
    q = 2 # half-width of local smoothing interval
    order = 3# order of fitting 2D spline curve
    # convert to logarithms
    lrho = np.log(rho)
    leta = np.log(eta)
    # filter points on Lcurve based on solution norm     
    lr = len(lrho)
    
    # smoothing 
    slrho = np.copy(lrho) 
    sleta = np.copy(leta)
    # for all interior points  q, lr-1-q, perform local smoothing 
    # with polynomial degree deg to k-q:k+q
    v = np.arange(-q, q+1)
    A = np.zeros((2*q + 1, deg + 1))
    A[:, 0] = np.ones((2*q + 1,))
    for j in range(1, deg + 1):
        A[:,j] = A[:,j-1] * v
    Ainv = np.linalg.pinv(A)
    for k in range(q, lr - q):
        slrho[k] = Ainv[0].dot(lrho[k + v])
        sleta[k] = Ainv[0].dot(leta[k + v])
    
    # fit a 2D spline curve to the smoothed discrete L-curve
    tck, u = interpolate.splprep([slrho.ravel(), sleta.ravel()], s = 0, k = order)
    spl_pts = 200 # number of pts on which to evaluate curvature 
    while(len(u)<spl_pts):
        # include u values at the middle of previous knots
        u = np.sort(np.hstack( [u , (u[1:] + u[:-1])/2]))
    xy = interpolate.splev(u, tck, der = 0)
    D = interpolate.splev(u, tck, der = 1)
    DD = interpolate.splev(u, tck, der = 2)
    # compute the corner of the discretized spline curve 
    # via max curvature
    k1 = D[0] * DD[1] - DD[0]*D[1] 
    k2 = (D[0]**2 + D[1]**2)**(1.5)
    kappa = safedivision(-k1, k2)
    # cut indices where monotonicity is not satisfied 
    id_max = len(kappa)
    min1 = np.where(D[1] < 0 )[0]
    min1 = min1[0] if len(min1) else id_max
    min2 = np.where(D[0] > 0 )[0]
    min2 = min2[0] if len(min2) else id_max
    cut_id = min( min1, min2 )  
    
    ikmax = np.argmax(kappa[:cut_id])
    x_corner, y_corner = interpolate.splev(u[ikmax], tck)
    assert x_corner == xy[0][ikmax]
    assert y_corner == xy[1][ikmax]
    kmax = kappa[ikmax]
    if kmax < 0:
            # then just take the smallest regularization parameter
            cid = -1
    else: 
    
            index = np.where(np.logical_and(lrho < x_corner,leta<y_corner))
            if len(index[0]) > 0:
                # restrict search values
                cid =  np.argmin((lrho[index] - x_corner)**2 + (leta[index]  - y_corner)**2)
                cid = index[0][cid]
            else: 
                # brute force search 
                cid = np.argmin((lrho - x_corner)**2 + (leta - y_corner)**2)
    if plot: 
        plt.figure() 
        plt.plot(xy[0][:cut_id], kappa[:cut_id], marker = '.')
        plt.scatter( xy[0][ikmax], kappa[ikmax], marker='v', color ='r')
        plt.xlabel('norm of residu ')
        plt.ylabel('curvature')
        plt.title('curvature')
        plt.show()
        plt.figure()
        plt.plot(xy[0][:cut_id], xy[1][:cut_id], '.', color = 'red', label= 'spline')
        plt.plot(lrho,leta, '.', label = 'logcoord')
        plt.scatter(lrho[cid], leta[cid], marker = 'x', color = 'black')
        plt.legend()
        plt.xlabel('norm of residu')
        plt.ylabel('norm of solution')
        plt.title('L - curve') 
        plt.show()
    return cid




from setuptools import setup

setup(
    name="rpcfit",
    version="0.9.9",
    author="Roland Akiki, Roger Marí",
    description="Robust Rational Polynomial Camera Modelling for SAR and Pushbroom Imaging",
    url="https://github.com/centreborelli/rpcfit",
    license="BSD",
    install_requires=[
        "numpy",
        "scipy",
        "matplotlib",
        "rasterio",
        "rpcm",
    ],
    packages=["rpcfit"],
    classifiers=[
        "License :: OSI Approved :: BSD License",
    ],
)
